from __future__ import division, print_function, absolute_import
import numpy as np
from ..math import gauss

from .a1sB_monomer import x0lambda_eval
from .monomer_aux import J_lam, I_lam

from .ares import ares, dares_drho, d2ares_drho, dares_dx, dares_dxrho

from .ideal import aideal, daideal_drho, d2aideal_drho, daideal_dx
from .ideal import daideal_dxrho

from .association_aux import association_config
from .association_aux import association_solver, association_check
from .polarGV import aij, bij, cij

from .density_solver import density_topliss, density_newton
from ..constants import kb, Na

from ..vrmie_pure import saftvrmie_pure

R = Na * kb


def U_mie(r, c, eps, lambda_r, lambda_a):
    u = c * eps * (np.power.outer(r, lambda_r) - np.power.outer(r, lambda_a))
    return u


# Second perturbation
phi16 = np.array([[7.5365557, -37.60463, 71.745953, -46.83552, -2.467982,
                 -0.50272, 8.0956883],
                 [-359.44, 1825.6, -3168.0, 1884.2, -0.82376, -3.1935, 3.7090],
                 [1550.9, -5070.1, 6534.6, -3288.7, -2.7171, 2.0883, 0],
                 [-1.19932, 9.063632, -17.9482, 11.34027, 20.52142, -56.6377,
                 40.53683],
                 [-1911.28, 21390.175, -51320.7, 37064.54, 1103.742, -3264.61,
                 2556.181],
                 [9236.9, -129430., 357230., -315530., 1390.2, -4518.2,
                 4241.6]])


nfi = np.arange(0, 7)
nfi_num = nfi[:4]
nfi_den = nfi[4:]


# Equation A26
def fi(alphaij, i):
    phi = phi16[i-1]
    num = np.tensordot(np.power.outer(alphaij, nfi_num), phi[nfi_num],
                       axes=(-1, -1))
    den = 1 + np.tensordot(np.power.outer(alphaij, nfi_den - 3), phi[nfi_den],
                           axes=(-1, -1))
    return num/den


class saftvrmie_mix():
    '''
    SAFT-VR-Mie EoS for mixtures Object

    This object have implemeted methods for phase equilibrium
    as for interfacial properties calculations.

    Parameters
    ----------
    mixture : object
        mixture created with mixture class

    Attributes
    ----------
    nc: integrer
        number of component in the mixture
    ms: array_like
        number of chain segments
    sigma: array_like
        size parameter of Mie potential [m]
    eps: array_like
        well-depth of Mie potential [J]
    la: array_like
        attractive exponent for Mie potential
    lr: array_like
        repulsive exponent for Mie potential

    ring: geometric parameter for ring molecules
          (see Langmuir 2017, 33, 11518-11529, Table I.)

    sigmaij: array_like
        size parameters matrix for Mie potential [m]
    epsij: array_like
        well-depth energy matrix of Mie potential [J]
    laij: array_like
        attractive exponent matrix for Mie potential
    lrij: array_like
        repulsive exponent matrix for Mie potential
    alphaij: array_like
        matrix of alpha van der waals constant

    eABij: array_like
        association energy matrix [J]
    rcij: array_like
        association range matrix [m]
    rdij: array_like
        association site position matrix [m]
    sites: list
        triplet of number of association sites [B, P, N]

    mupol: array_like
        dipolar moment [Debye]
    npol: array_like
        number of dipolar sites

    cii : array_like
        influence factor for SGT [J m^5 / mol^2]
    cij : array_like
        cross influence parameter matrix for SGT [J m^5 / mol^2]
    beta: array_like
        correction to cross influence parameter matrix

    secondorder: bool
        bool to indicate if composition derivatives of fugacity coefficient
        are available
    secondordersgt: bool
        bool to indicate if composition derivatives of chemical potential
        are available


    Methods
    -------
    cii_correlation : correlates the influence parameter of the fluid
    diameter : computes the diameter at given temperature
    temperature_aux : computes temperature depedent parameters of the fluid
    ares: computes the residual dimentionless Helmholtz free energy
    dares_drho: computes the residual dimentionless Helmholtz free energy
                and its density first density derivative
    d2ares_drho: computes the residual dimentionless Helmholtz free energy
                 and its density first and second density derivatives
    dares_dx: computes the residual dimentionless Helmholtz free energy and
              its composition derivatives
    dares_dxrho: computes the residual dimentionless Helmholtz free energy
                 and its composition and density derivatives
    afcn: computes total Helmholtz energy
    dafcn_drho: computes total Helmholtz energy and its density derivative
    d2afcn_drho: computes total Helmholtz energy and it density derivatives
    dafcn_dx: computes total Helmholtz energy and its composition derivative
    dafcn_dxrho:computes total Helmholtz energy and its composition and
                density derivative
    density: computes the density of the fluid
    pressure: computes the pressure
    dP_drho: computes pressure and its density derivative
    logfugmix: computes the fugacity coefficient of the mixture
    logfugef: computes the effective fugacity coefficients of the components
              in the mixture
    a0ad: computes dimensionless Helmholtz density energy
    muad: computes dimensionless chemical potential
    dmuad: computes the dimensionless chemical potential and its derivatives
    dOm : computes dimensionless Thermodynamic Grand Potential
    ci :  computes influence parameters matrix for SGT
    sgt_adim : computes dimensionless factors for SGT
    beta_sgt: method for setting beta correction used in SGT

    association_solver: computes the fraction of non-bonded sites
    association_check: checks the fraction of non-bonded sites solution

    EntropyR : computes the residual entropy of the fluid
    EnthalpyR : computes the residual enthalpy of the fluid
    CvR : computes the residual isochoric heat capacity
    CpR : computes the residual heat capacity
    speed_sound : computes the speed of sound

    Auxiliar methods (computed using temperature_aux output list)
    -------------------------------------------------------------
    density_aux : computes density
    afcn_aux : computes afcn
    dafcn_aux : computes dafcn_drho
    d2afcn_aux : computes d2afcn_drho
    pressure_aux : computes pressure
    dP_drho_aux : computes dP_drho
    logfugmix_aux : computes logfug
    a0ad_aux : compute a0ad
    muad_aux : computes muad
    dmuad_aux : computes dmuad
    dOm_aux : computes dOm
    '''

    def __init__(self, mixture, compute_critical=True):

        self.mixture = mixture
        self.Mw = np.asarray(mixture.Mw)

        # Pure component parameters
        self.lr = np.asarray(mixture.lr)
        self.la = np.asarray(mixture.la)
        self.lar = self.lr + self.la
        self.sigma = np.asarray(mixture.sigma)
        self.eps = np.asarray(mixture.eps)
        self.ms = np.asarray(mixture.ms)
        self.ring = np.asarray(mixture.ring)

        self.nc = mixture.nc

        sigma3 = self.sigma**3
        # Eq A45
        self.sigmaij = np.add.outer(self.sigma, self.sigma) / 2
        self.sigmaij3 = self.sigmaij**3
        # Eq A51

        self.epsij = np.multiply.outer(sigma3, sigma3)
        self.epsij *= np.multiply.outer(self.eps, self.eps)
        self.epsij **= 0.5
        self.epsij /= self.sigmaij3
        '''
        if hasattr(mixture, 'KIJsaft'):
            kij0 = mixture.KIJ0saft
        else:
            kij0 = np.zeros([mixture.nc, mixture.nc])
        # self.epsij *= (1-kij)
        '''

        # kij matrix: kij = kij0 + kij1 T + kij2 T^2 + kij3 / T
        self.KIJ0saft = mixture.KIJ0saft
        self.KIJ1saft = mixture.KIJ1saft
        self.KIJ2saft = mixture.KIJ2saft
        self.KIJ3saft = mixture.KIJ3saft

        # Eq A48
        self.lrij = np.sqrt(np.multiply.outer(self.lr-3., self.lr-3.)) + 3
        self.laij = np.sqrt(np.multiply.outer(self.la-3., self.la-3.)) + 3
        self.larij = self.lrij + self.laij

        # Eq A3
        dif_cij = self.lrij - self.laij
        self.Cij = self.lrij/dif_cij*(self.lrij/self.laij)**(self.laij/dif_cij)
        # Eq A24
        self.alphaij = self.Cij * (1./(self.laij - 3.) - 1./(self.lrij - 3.))
        self.Cij2 = self.Cij**2

        self.diag_index = np.diag_indices(self.nc)
        self.C = self.Cij[self.diag_index]
        self.C2 = self.Cij[self.diag_index]**2
        self.alpha = self.alphaij[self.diag_index]

        # For Second perturbation
        self.f1 = fi(self.alphaij, 1)
        self.f2 = fi(self.alphaij, 2)
        self.f3 = fi(self.alphaij, 3)
        self.f4 = fi(self.alphaij, 4)
        self.f5 = fi(self.alphaij, 5)
        self.f6 = fi(self.alphaij, 6)

        # Eq A18
        c_matrix = np.array([[0.81096, 1.7888, -37.578, 92.284],
                            [1.0205, -19.341, 151.26, -463.5],
                            [-1.9057, 22.845, -228.14, 973.92],
                            [1.0885, -6.1962, 106.98, -677.64]])

        lrij_power = np.power.outer(self.lrij, [0, -1, -2, -3])
        self.cctes_lrij = np.tensordot(lrij_power, c_matrix, axes=(-1, -1))

        laij_power = np.power.outer(self.laij, [0, -1, -2, -3])
        self.cctes_laij = np.tensordot(laij_power, c_matrix, axes=(-1, -1))

        lrij_2power = np.power.outer(2*self.lrij, [0, -1, -2, -3])
        self.cctes_2lrij = np.tensordot(lrij_2power, c_matrix, axes=(-1, -1))

        laij_2power = np.power.outer(2*self.laij, [0, -1, -2, -3])
        self.cctes_2laij = np.tensordot(laij_2power, c_matrix, axes=(-1, -1))

        larij_power = np.power.outer(self.larij, [0, -1, -2, -3])
        self.cctes_larij = np.tensordot(larij_power, c_matrix, axes=(-1, -1))

        # Monomer necessary term
        self.lambdasij = (self.laij, self.lrij, self.larij)
        self.cctesij = (self.cctes_laij, self.cctes_lrij,
                        self.cctes_2laij, self.cctes_2lrij, self.cctes_larij)

        # Chain neccesary term
        self.cctes_la = self.cctes_laij[self.diag_index]
        self.cctes_lr = self.cctes_lrij[self.diag_index]
        self.cctes_lar = self.cctes_larij[self.diag_index]
        self.cctes_2la = self.cctes_2laij[self.diag_index]
        self.cctes_2lr = self.cctes_2lrij[self.diag_index]

        self.lambdas = (self.la, self.lr, self.lar)
        self.cctes = (self.cctes_la, self.cctes_lr, self.cctes_2la,
                      self.cctes_2lr, self.cctes_lar)

        # For diameter calculation
        roots, weights = gauss(30)
        self.roots = roots
        self.weights = weights
        self.umie = U_mie(1./roots, self.C, self.eps, self.lr, self.la)

        self.dxhi00_drho = np.pi / 6

        # association config
        self.eABij = np.sqrt(np.multiply.outer(mixture.eAB, mixture.eAB))
        # lij matrix: lij = lij0 + lij1 T + lij2 T^2 + lij3 / T
        self.LIJ0saft = mixture.LIJ0saft
        self.LIJ1saft = mixture.LIJ1saft
        self.LIJ2saft = mixture.LIJ2saft
        self.LIJ3saft = mixture.LIJ3saft
        '''
        if hasattr(mixture, 'LIJsaft'):
            lij = mixture.LIJsaft
        else:
            lij = np.zeros([mixture.nc, mixture.nc])
        self.eABij *= (1.-lij)
        '''
        self.rcij = np.add.outer(mixture.rc, mixture.rc)/2
        self.rdij = np.add.outer(mixture.rd, mixture.rd)/2

        S, DIJ, compindex, indexabij, indexab, nsites, \
        dxjdx, diagasso = association_config(mixture.sitesmix, self)
        assoc_bool = nsites != 0
        self.assoc_bool = assoc_bool
        if assoc_bool:
            self.S = S
            self.DIJ = DIJ
            self.compindex = compindex
            self.indexab = indexab
            self.indexabij = indexabij
            self.nsites = nsites
            self.dxjdx = dxjdx
            self.diagasso = diagasso

        self.secondorder = False
        self.secondordersgt = True

        # Polar Contribution
        self.mupol = np.asarray(mixture.mupol)
        self.npol = np.asarray(mixture.npol)
        polar_bool = np.any(self.npol != 0)
        self.polar_bool = polar_bool
        if polar_bool:
            ms = self.ms
            self.mpol = ms * (ms < 2) + 2 * (ms >= 2)
            msij = np.sqrt(np.outer(ms, ms))
            msijk = np.cbrt(np.multiply.outer(ms, np.outer(ms, ms)))
            msij[msij > 2.] = 2.
            msijk[msijk > 2.] = 2.

            aux1 = np.array([np.ones_like(msij), (msij-1)/msij,
                            (msij-1)/msij * (msij-2)/msij])
            aux2 = np.array([np.ones_like(msijk), (msijk-1)/msijk,
                            (msijk-1)/msijk * (msijk-2)/msijk])
            self.anij = np.tensordot(aij, aux1, axes=(1))
            self.bnij = np.tensordot(bij, aux1, axes=(1))
            self.cnij = np.tensordot(cij, aux2, axes=(1))

            sigmaij = self.sigmaij
            mult = np.multiply.outer(sigmaij, sigmaij)
            listpolar = np.arange(self.nc)
            self.sigmaijk3 = mult[listpolar, :, listpolar] * sigmaij
            self.sigma3 = self.sigma**3
            # 1 D = 3.33564e-30 C * m
            # 1 C^2 = 9e9 N m^2
            cte = (3.33564e-30)**2 * (9e9)
            self.mupolad2 = self.mupol**2*cte/(self.ms*self.eps*self.sigma3)

        self.cii = mixture.cii
        # self.cij = np.sqrt(np.outer(self.cii, self.cii))
        # self.beta = np.zeros([self.nc, self.nc])
        self.beta0 = np.zeros([self.nc, self.nc])
        self.beta1 = np.zeros([self.nc, self.nc])
        self.beta2 = np.zeros([self.nc, self.nc])
        self.beta3 = np.zeros([self.nc, self.nc])

        # creating pure fluid's eos
        pure_eos = []
        for i, component in enumerate(self.mixture.components):
            model = saftvrmie_pure(component, compute_critical=compute_critical)
            pure_eos.append(model)
        self.pure_eos = pure_eos

    def set_induced_asso(self, selfasso, inducedasso, rcij, eABij=None):
        r"""
        This method modifies the eos.rcij and eos.eABij attributes to set
        induced association (solvation). You need to provide the fitted rcij
        distance in Amstrong (used to compute the association volume).
        If no induced association energy is supplied, by default the half of
        the self-association energy of the self-associative component is used.

        Parameters
        ----------
        selfasso : int
            index of  self associating component.
        inducedasso : int
            index of induced associating component.
        rcij : float
            rcij distance (fitted) [Amstrong]
        eABij : float, optional
            cross-association energy [K]. Default to eAB(self-associating)/2.

        """
        type_self = type(selfasso) == int
        typej_ind = type(inducedasso) == int

        nc = self.nc
        nc_self = 0 <= selfasso <= (nc - 1)
        nc_ind = 0 <= inducedasso <= (nc - 1)

        self_induced = selfasso != inducedasso

        if (not nc_self) or (not nc_ind):
            raise Exception('Index selfasso or inducedasso bigger than (nc-1)')
        if not self_induced:
            raise Exception('Cannot set induced association for selfasso=inducedasso')

        if type_self and typej_ind and nc_self and nc_ind:
            self.rcij[selfasso, inducedasso] = rcij * 1e-10  # rcij in m
            self.rcij[inducedasso, selfasso] = rcij * 1e-10  # rcij in m
            if eABij is None:
                self.eABij[selfasso, inducedasso] = self.eABij[selfasso, selfasso] / 2
                self.eABij[inducedasso, selfasso] = self.eABij[selfasso, selfasso] / 2
            else:
                self.eABij[selfasso, inducedasso] = eABij * kb  # eABij in J
                self.eABij[inducedasso, selfasso] = eABij * kb  # eABij in J

    def cii_correlation(self, overwrite=False):
        """
        cii_corelation()

        Method that computes the influence parameter of coarsed-grained
        molecules

        AIChE Journal, 62, 5, 1781-1794 (2016)
        Eq. (23)

        Parameters
        ----------

        overwrite : bool
            If true it will overwrite the actual influence parameter.

        Returns
        -------
        cii : array_like
            correlated influence parameter [J m^5 / mol^2]
        """
        cii = self.ms * (0.12008072630855947 + 2.2197907527439655 * self.alpha)
        cii *= np.sqrt(Na**2 * self.eps * self.sigma**5)
        cii **= 2

        cii = np.array(cii).reshape([self.nc, 1])
        if overwrite:
            self.cii = cii
            # self.cij = np.sqrt(np.outer(self.cii, self.cii))
        return cii

    def diameter(self, beta):
        """
        diameter(beta)

        Method that computes the diameter of the fluids at given
        beta = 1 / kb T

        Journal of Chemical Physics, 139(15), 1–37 (2013)
        Eq. (7)

        Parameters
        ----------

        beta : float
            Boltzmann's factor: beta = 1 / kb T [1/J]

        Returns
        -------
        d : array_like
            computed diameters [m]
        """
        # umie = U_mie(1/roots, c, eps, lambda_r, lambda_a)
        integrer = np.exp(-beta * self.umie)
        d = self.sigma * (1. - np.matmul(self.weights, integrer))
        return d

    def temperature_aux(self, T):
        """
        temperature_aux(T)

        Method that computes temperature dependent parameters.
        It returns the following list:

        temp_aux = [beta, beta2, beta3, dii, dij, x0, x0i, di03, dij3,
                    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij,
                    beps, beps2, a1vdw_cte, x0i_matrix, tetha,
                    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii,
                    Fab, aux_dii, aux_dii2, Kab, epsa, epsija]

        Journal of Chemical Physics, 139(15), 1–37 (2013)

        beta: Boltzmann's factor [1/J]
        beta2: beta**2  [1/J^2]
        beta3: beta**3  [1/J^3]
        dii: computed diameters [m] (Eq A9)
        dij: diameters matrix [m] (Eq A46)
        x0: sigmaij/dij [Adim] (Below Eq. A11)
        x0i: sigma/dii [Adim]
        di03: dii^3 [m^3]
        dij3: dij^3 [m^3]
        I_lambdasij: (Eq A14)
        J_lambdasij: (Eq A15)
        a1vdw_cteij: -12 * epsij * dij3
        a1vdwij: tuple using a1vdw_cteij with different lamda_ij exponents
        beps: beta*eps [Adim]
        beps2: betps**2 [Adim]
        a1vdw_cte: -12 * eps * dii3
        x0i_matrix: np.array([x0i**0, x0i**1, x0i**2, x0i**3]) used in (Eq A29)
        tetha: exp(beta*eps)-1 [Adim] (Below Eq. A37)
        x0_a1, x0_a2, x0_g1, x0_g2: used to compute a1ij, a2ij
        x0_a1ii, x0_a2ii: diagonal of x0_a1, x0_a2
        Fab: association strength [Adim] (Eq. A41)
        aux_dii, aux_dii2: auxiliar variables used in association contribution
        Kab: association volume computed with  Eq. A43
        epsa: eps / kb / T [Adim]
        epsija: epsij / kb / T [Adim]

        Parameters
        ----------

        T : float
            Absolute temperature [K]

        Returns
        -------
        temp_aux : list
             list of computed parameters
        """
        diag_index = self.diag_index

        beta = 1 / (kb * T)
        beta2 = beta**2
        beta3 = beta**3

        dii = self.diameter(beta)
        # Eq A46
        dij = np.add.outer(dii, dii) / 2
        x0 = self.sigmaij / dij
        x0i = x0[diag_index]

        # Defining xhi0 without x dependence
        di03 = np.power.outer(dii, np.arange(4))
        dij3 = dij**3

        # used in a1, a2, g1 and g2
        out = x0lambda_eval(x0, self.la, self.lr, self.lar, self.laij,
                            self.lrij, self.larij, diag_index)
        x0_a1, x0_a2, x0_g1, x0_g2 = out
        x0_a1ii = x0_a1[:, diag_index[0], diag_index[1]]
        x0_a2ii = x0_a2[:, diag_index[0], diag_index[1]]

        # I and J used B term
        I_la = I_lam(x0, self.laij)
        I_lr = I_lam(x0, self.lrij)
        I_2la = I_lam(x0, 2*self.laij)
        I_2lr = I_lam(x0, 2*self.lrij)
        I_lar = I_lam(x0, self.larij)
        I_lambdasij = (I_la, I_lr, I_2la, I_2lr, I_lar)

        J_la = J_lam(x0, self.laij)
        J_lr = J_lam(x0, self.lrij)
        J_2la = J_lam(x0, 2*self.laij)
        J_2lr = J_lam(x0, 2*self.lrij)
        J_lar = J_lam(x0, self.larij)
        J_lambdasij = (J_la, J_lr, J_2la, J_2lr, J_lar)

        # computing temperature dependent kij and epsij
        kij = self.KIJ0saft + self.KIJ1saft*T + self.KIJ2saft*T**2
        kij += self.KIJ3saft/T
        epsij = self.epsij*(1.-kij)

        # Monomer necessary terms
        # a1vdw_cteij = -12 * self.epsij * dij3
        a1vdw_cteij = -12 * epsij * dij3

        a1vdw_laij = a1vdw_cteij / (self.laij - 3)
        a1vdw_lrij = a1vdw_cteij / (self.lrij - 3)
        a1vdw_2laij = a1vdw_cteij / (2*self.laij - 3)
        a1vdw_2lrij = a1vdw_cteij / (2*self.lrij - 3)
        a1vdw_larij = a1vdw_cteij / (self.larij - 3)
        a1vdwij = (a1vdw_laij, a1vdw_lrij, a1vdw_2laij, a1vdw_2lrij,
                   a1vdw_larij)

        # Chain necessary terms
        a1vdw_cte = a1vdw_cteij[diag_index]

        x0i_matrix = np.array([x0i**0, x0i, x0i**2, x0i**3])

        beps = beta * self.eps
        beps2 = beps**2
        # tetha = np.exp(beta * self.eps) - 1.
        tetha = np.exp(beps) - 1.

        # computing temperature depedent lij and epsABij
        lij = self.LIJ0saft + self.LIJ1saft*T + self.LIJ2saft*T**2
        lij += self.LIJ3saft/T
        eABij = self.eABij*(1.-lij)
        # For associating mixtures
        # Fab = np.exp(beta * self.eABij) - 1.
        Fab = np.exp(beta * eABij) - 1.
        aux_dii = np.multiply.outer(dii, dii)/np.add.outer(dii, dii)
        aux_dii2 = aux_dii**2
        dij2 = dij**2

        rcij = self.rcij
        rcij2 = self.rcij**2
        rcij3 = rcij2*rcij
        rdij = self.rdij
        rdij2 = self.rdij**2
        rdij3 = rdij2*rdij

        Kab = np.log((rcij + 2*rdij)/dij)
        Kab *= 6*rcij3 + 18 * rcij2*rdij - 24 * rdij3
        aux1 = (rcij + 2 * rdij - dij)
        aux2 = (22*rdij2 - 5*rcij*rdij - 7*rdij*dij - 8*rcij2+rcij*dij+dij2)
        Kab += aux1 * aux2
        Kab /= (72*rdij2 * self.sigmaij3)
        Kab *= 4 * np.pi * dij2

        # For Polar mixtures
        epsa = beps
        # epsija = self.epsij * beta
        epsija = epsij * beta

        temp_aux = [beta, beta2, beta3, epsij, dii, dij, x0, x0i, di03, dij3,
                    I_lambdasij, J_lambdasij, a1vdw_cteij, a1vdwij,
                    beps, beps2, a1vdw_cte, x0i_matrix, tetha,
                    x0_a1, x0_a2, x0_g1, x0_g2, x0_a1ii, x0_a2ii,
                    Fab, aux_dii, aux_dii2, Kab, epsa, epsija]
        return temp_aux

    def association_solver(self, x, rho, T, Xass0=None):
        """
        association_solver(x, rho, T, Xass0)

        Method that computes the fraction of non-bonded sites.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molar density [mol/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        Returns
        -------
        Xass: array_like
            fraction of non-bonded sites. (None if the mixture is non-associating)
        """
        if self.assoc_bool:
            temp_aux = self.temperature_aux(T)
            rhom = rho * Na
            Xass = association_solver(self, x, rhom, temp_aux, Xass0)
        else:
            Xass = None
        return Xass

    def association_check(self, x, rho, T, Xass):
        """
        association_check(x, rho, T, Xass0)

        Method that checks if the association sites are consistent.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molar density [mol/m3]
        T: float
            absolute temperature [K]
        Xass: array
            Initial guess for the calculation of fraction of non-bonded sites
        Returns
        -------
        of: array_like
            objective function that should be zero if the association sites
            are consistent. (Zero if the mixture is non-associating)
        """
        if self.assoc_bool:
            temp_aux = self.temperature_aux(T)
            rhom = rho * Na
            fo = association_check(self, x, rhom, temp_aux, Xass)
        else:
            fo = 0.
        return fo

    def ares(self, x, rho, T, Xass0=None):
        """
        ares(x, rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the mixture.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           residual dimentionless Helmholtz free energy [Adim]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = ares(self, x, rho, temp_aux, Xass0)
        return a

    def dares_drho(self, x, rho, T, Xass0=None):
        """
        dares_drho(x, rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the mixture
        and its first density derivative.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           residual dimentionless Helmholtz free energy [Adim, m^3]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = dares_drho(self, x, rho, temp_aux, Xass0)
        return a

    def d2ares_drho(self, x, rho, T, Xass0=None):
        """
        d2ares_drho(x, rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the mixture
        and its first and second density derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           residual dimentionless Helmholtz free energy [Adim, m^3, m^6]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = d2ares_drho(self, x, rho, temp_aux, Xass0)
        return a

    def dares_dx(self, x, rho, T, Xass0=None):
        """
        dares_dx(x, rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the mixture
        and its composition derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           residual dimentionless Helmholtz free energy [Adim]
        ax: array_like
           composition derivatives of residual dimentionless Helmholtz
           free energy [Adim]
        """
        temp_aux = self.temperature_aux(T)
        a, ax, Xass = dares_dx(self, x, rho, temp_aux, Xass0)
        return a, ax

    def dares_dxrho(self, x, rho, T, Xass0=None):
        """
        dares_dx(x, rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the mixture
        and its density and composition derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           residual dimentionless Helmholtz free energy [Adim, m^3]
        ax: array_like
           composition derivatives of residual dimentionless Helmholtz
           free energy [Adim]
        """
        temp_aux = self.temperature_aux(T)
        a, ax, Xass = dares_dxrho(self, x, rho, temp_aux, Xass0)
        return a, ax

    def afcn_aux(self, x, rho, temp_aux, Xass0=None):
        """
        afcn_aux(x, rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the mixture.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           total Helmholtz free energy [J/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        a, Xass = ares(self, x, rho, temp_aux, Xass0)
        a += aideal(x, rho, beta)
        a *= (Na/beta)
        return a, Xass

    def dafcn_drho_aux(self, x, rho, temp_aux, Xass0=None):
        """
        dafcn_drho_aux(x, rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its first density derivative.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           toal Helmholtz free energy [J/mol, J m^3/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        a, Xass = dares_drho(self, x, rho, temp_aux, Xass0)
        a += daideal_drho(x, rho, beta)
        a *= (Na/beta)
        return a, Xass

    def d2afcn_drho_aux(self, x, rho, temp_aux, Xass0=None):
        """
        d2afcn_drho_aux(x, rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its first and second density derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           toal Helmholtz free energy [J/mol, J m^3/mol, J m^6/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        a, Xass = d2ares_drho(self, x, rho, temp_aux, Xass0)
        a += d2aideal_drho(x, rho, beta)
        a *= (Na/beta)
        return a, Xass

    def dafcn_dx_aux(self, x, rho, temp_aux, Xass0=None):
        """
        dafcn_dx_aux(x, rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its composition derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           total dimentionless Helmholtz free energy [J/mol]
        ax: array_like
           composition derivatives of total dimentionless Helmholtz
           free energy [J/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        ar, aresx, Xass = dares_dx(self, x, rho, temp_aux, Xass0)
        aideal, aidealx = daideal_dx(x, rho, beta)
        a = (ar + aideal)
        a *= (Na/beta)
        ax = (aresx + aidealx)
        ax *= (Na/beta)
        return a, ax, Xass

    def dafcn_dxrho_aux(self, x, rho, temp_aux, Xass0=None):
        """
        dafcn_dxrho_aux(x, rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its composition a density derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           total dimentionless Helmholtz free energy [J/mol, J m^3/mol]
        ax: array_like
           composition derivatives of total dimentionless Helmholtz
           free energy [J/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        ar, aresx, Xass = dares_dxrho(self, x, rho, temp_aux, Xass0)
        aideal, aidealx = daideal_dxrho(x, rho, beta)
        a = (ar + aideal)
        a *= (Na/beta)
        ax = (aresx + aidealx)
        ax *= (Na/beta)
        return a, ax, Xass

    def afcn(self, x, rho, T, Xass0=None):
        """
        afcn(x, rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the mixture.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           total Helmholtz free energy [J/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = self.afcn_aux(x, rho, temp_aux, Xass0)
        return a

    def dafcn_drho(self, x, rho, T, Xass0=None):
        """
        dafcn_drho(x, rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its first density derivative.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           toal Helmholtz free energy [J/mol, J m^3/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = self.dafcn_drho_aux(x, rho, temp_aux, Xass0)
        return a

    def d2afcn_drho(self, x, rho, T, Xass0=None):
        """
        d2afcn_drho(x, rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its first and second density derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           toal Helmholtz free energy [J/mol, J m^3/mol, J m^6/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = self.d2afcn_drho_aux(x, rho, temp_aux, Xass0)
        return a

    def dafcn_dx(self, x, rho, T, Xass0=None):
        """
        dafcn_dx(x, rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its composition derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           total dimentionless Helmholtz free energy [J/mol]
        ax: array_like
           composition derivatives of total dimentionless Helmholtz
           free energy [J/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, ax, Xass = self.dafcn_dx_aux(x, rho, temp_aux, Xass0)
        return a, ax

    def dafcn_dxrho(self, x, rho, T, Xass0=None):
        """
        dafcn_dxrho(x, rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the mixture
        and its composition a density derivatives.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           total dimentionless Helmholtz free energy [J/mol, J m^3/mol]
        ax: array_like
           composition derivatives of total dimentionless Helmholtz
           free energy [J/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, ax, Xass = self.dafcn_dxrho_aux(x, rho, temp_aux, Xass0)
        return a, ax

    def density_aux(self, x, temp_aux, P, state, rho0=None, Xass0=None):
        """
        density_aux(x, temp_aux, P, state)
        Method that computes the density of the mixture at given composition,
        temperature, pressure and aggregation state.

        Parameters
        ----------
        x: array_like
            molar fraction array
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapor phase
        rho0 : float, optional
            initial guess to compute density root [mol/m^3]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        density: float
            density [mol/m^3]
        Xass : array
            computed fraction of nonbonded sites
        """
        if rho0 is None:
            rho, Xass = density_topliss(state, x, temp_aux, P, Xass0, self)
        else:
            rho, Xass = density_newton(rho0, x, temp_aux, P, Xass0, self)
        return rho, Xass

    def density(self, x, T, P, state, rho0=None, Xass0=None):
        """
        density(x, T, P, state)
        Method that computes the density of the mixture at given composition,
        temperature, pressure and aggregation state.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapor phase
        rho0 : float, optional
            initial guess to compute density root [mol/m^3]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        density: float
            density [mol/m^3]
        """
        temp_aux = self.temperature_aux(T)
        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)
        return rho

    def pressure_aux(self, x, rho, temp_aux, Xass0=None):
        """
        pressure_aux(x, rho, temp_aux, Xass0)

        Method that computes the pressure at given composition,
        density [mol/m3] and temperature [K]

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            density [mol/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        Xass : array
            computed fraction of nonbonded sites
        """
        rhomolecular = Na * rho
        da, Xass = self.dafcn_drho_aux(x, rhomolecular, temp_aux, Xass0)
        afcn, dafcn = da
        Psaft = rhomolecular**2 * dafcn / Na
        return Psaft, Xass

    def dP_drho_aux(self, x, rho, temp_aux, Xass0=None):
        """
        dP_drho_aux(rho, temp_aux, Xass0)

        Method that computes the pressure and its density derivative at given
        composition, density [mol/m3] and temperature [K]

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            density [mol/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        dP: float
            derivate of pressure respect density [Pa m^3 / mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        rhomolecular = Na * rho
        da, Xass = self.d2afcn_drho_aux(x, rhomolecular, temp_aux, Xass0)
        afcn, dafcn, d2afcn = da
        Psaft = rhomolecular**2 * dafcn / Na
        dPsaft = 2 * rhomolecular * dafcn + rhomolecular**2 * d2afcn
        return Psaft, dPsaft, Xass

    def pressure(self, x, rho, T, Xass0=None):
        """
        pressure(x, rho, T, Xass0)

        Method that computes the pressure at given composition,
        density [mol/m3] and temperature [K]

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            density [mol/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        """
        temp_aux = self.temperature_aux(T)
        Psaft, Xass = self.pressure_aux(x, rho, temp_aux, Xass0)
        return Psaft

    def dP_drho(self, x, rho, T, Xass0=None):
        """
        dP_drho(rho, T, Xass0)

        Method that computes the pressure and its density derivative at given
        composition, density [mol/m3] and temperature [K]

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho: float
            density [mol/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        dP: float
            derivate of pressure respect density [Pa m^3 / mol]
        """
        temp_aux = self.temperature_aux(T)
        Psaft, dPsaft, Xass = self.dP_drho_aux(x, rho, temp_aux, Xass0)
        return Psaft, dPsaft

    def logfugmix_aux(self, x, temp_aux, P, state, v0=None, Xass0=None):
        """
        logfugmix_aux(x, temp_aux, P, state, v0, Xass0)

        Method that computes the fugacity coefficient of the mixture at given
        composition, temperature and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        P: float
            pressure [Pa]
        state: string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        lnphi: float
            fugacity coefficient of the mixture
        v: float
            computed volume of the phase [m^3/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        RT = Na/beta

        if v0 is None:
            rho0 = None
        else:
            rho0 = 1./v0
        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)

        v = 1./rho
        rhomolecular = Na * rho
        ar, Xass = ares(self, x, rhomolecular, temp_aux, Xass)
        Z = P * v / RT
        lnphi = ar + (Z - 1.) - np.log(Z)
        return lnphi, v, Xass

    def logfugmix(self, x, T, P, state, v0=None, Xass0=None):
        """
        logfugmix(x, T, P, state, v0, Xass0)

        Method that computes the fugacity coefficient of the mixture at given
        composition, temperature and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T: float
            absolute temperature [K]
        P: float
            pressure [Pa]
        state: string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        lnphi: float
            fugacity coefficient of the mixture
        v: float
            computed volume of the phase [m^3/mol]
        """
        temp_aux = self.temperature_aux(T)
        lnphi, v, Xass = self.logfugmix_aux(x, temp_aux, P, state, v0, Xass0)
        return lnphi, v

    def logfugef_aux(self, x, temp_aux, P, state, v0=None, Xass0=None):
        """
        logfugef_aux(x, temp_aux, P, state, v0, Xass0)

        Method that computes the effective fugacity coefficient of the
        components in the mixture at given composition, temperature
        and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        P: float
            pressure [Pa]
        state: string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        lnphi: array_like
            effective fugacity coefficient of the components
        v: float
            computed volume of the phase [m^3/mol]
        Xass : array
            computed fraction of nonbonded sites
        """

        beta = temp_aux[0]
        RT = Na/beta

        if v0 is None:
            rho0 = None
        else:
            rho0 = 1./v0

        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)
        v = 1./rho
        rhomolecular = Na * rho
        ar, daresx, Xass = dares_dx(self, x, rhomolecular, temp_aux, Xass)
        Z = P * v / RT
        mures = ar + (Z - 1.) + daresx - np.dot(x, daresx)
        lnphi = mures - np.log(Z)
        return lnphi, v, Xass

    def logfugef(self, x, T, P, state, v0=None, Xass0=None):
        """
        logfugef(x, T, P, state, v0, Xass0)

        Method that computes the effective fugacity coefficient of the
        components in the mixture at given composition, temperature
        and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T: float
            absolute temperature [K]
        P: float
            pressure [Pa]
        state: string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        lnphi: array_like
            effective fugacity coefficient of the components
        v: float
            computed volume of the phase [m^3/mol]
        """
        temp_aux = self.temperature_aux(T)
        lnphi, v, Xass = self.logfugef_aux(x, temp_aux, P, state, v0, Xass0)
        return lnphi, v

    def a0ad_aux(self, rhoi, temp_aux, Xass0=None):
        """
        a0ad_aux(rhoi, temp_aux, Xass0)

        Method that computes the Helmholtz density energy divided by RT at
        given density vector and temperature.

        Parameters
        ----------

        rhoi : array_like
            density vector [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a0ad: float
            Helmholtz density energy divided by RT [mol/m^3]
        Xass : array
            computed fraction of nonbonded sites
        """

        rho = np.sum(rhoi)
        x = rhoi/rho
        rhomolecular = Na * rho
        beta = temp_aux[0]

        a, Xass = ares(self, x, rhomolecular, temp_aux, Xass0)
        a += aideal(x, rhomolecular, beta)

        a0 = a*rho

        return a0, Xass

    def a0ad(self, rhoi, T, Xass0=None):
        """
        a0ad(rhoi, T, Xass0)

        Method that computes the Helmholtz density energy divided by RT at
        given density vector and temperature.

        Parameters
        ----------

        rhoi : array_like
            density vector [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a0ad: float
            Helmholtz density energy divided by RT [mol/m^3]
        """
        temp_aux = self.temperature_aux(T)
        a0, Xass = self.a0ad_aux(rhoi, temp_aux, Xass0)
        return a0

    def muad_aux(self, rhoi, temp_aux, Xass0=None):
        """
        muad_aux(rhoi, temp_aux, Xass0)

        Method that computes the dimentionless chemical potential at given
        density vector and temperature.

        Parameters
        ----------
        rhoi : array_like
            density vector [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        muad: array_like
            chemical potential divided by RT [Adim]
        Xass : array
            computed fraction of nonbonded sites
        """
        rho = np.sum(rhoi)
        x = rhoi/rho
        rhom = Na * rho

        beta = temp_aux[0]
        ares, aresx, Xass = dares_dxrho(self, x, rhom, temp_aux, Xass0)
        aideal, aidealx = daideal_dxrho(x, rhom, beta)
        afcn, dafcn = (ares + aideal)
        ax = (aresx + aidealx)
        Z = dafcn * rhom
        mu = afcn + ax - np.dot(x, ax) + (Z)
        return mu, Xass

    def muad(self, rhoi, T, Xass0=None):
        """
        muad(rhoi, T, Xass0)

        Method that computes the dimentionless chemical potential at given
        density vector and temperature.

        Parameters
        ----------
        rhoi : array_like
            density vector [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        muad: array_like
            chemical potential divided by RT [Adim]
        """
        temp_aux = self.temperature_aux(T)
        mu, Xass = self.muad_aux(rhoi, temp_aux, Xass0)
        return mu

    def dmuad_aux(self, rhoi, temp_aux, Xass0=None):
        """
        dmuad_aux(rhoi, temp_aux, Xass0)

        Method that computes the chemical potential and its numerical
        derivative at given density vector and temperature.

        Parameters
        ----------
        rhoi : array_like
            density vector [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        muad: array_like
            chemical potential divided by RT [Adim]
        dmuad: array_like
            derivavites of the chemical potential respect to rhoi [m^3/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        nc = self.nc
        h = 1e-3
        diff = h * np.eye(nc)

        mu, Xass = self.muad_aux(rhoi, temp_aux, Xass0)

        arr = []
        for i in range(nc):
            muad1, _ = self.muad_aux(rhoi + diff[i], temp_aux, Xass)
            muad_1, _ = self.muad_aux(rhoi - diff[i], temp_aux, Xass)
            arr.append((muad1 - muad_1)/(2*h))

        dmu = np.column_stack(arr)
        return mu, dmu, Xass

    def dmuad(self, rhoi, T, Xass0=None):
        """
        dmuad(rhoi, T, Xass0)

        Method that computes the chemical potential and its numerical
        derivative at given density vector and temperature.

        Parameters
        ----------
        rhoi : array_like
            density vector [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        muad: array_like
            chemical potential divided by RT [Adim]
        dmuad: array_like
            derivavites of the chemical potential respect to rhoi [m^3/mol]
        """
        temp_aux = self.temperature_aux(T)
        mu, dmu, Xass = self.dmuad_aux(rhoi, temp_aux, Xass0)
        return mu, dmu

    def dOm_aux(self, rhoi, temp_aux, mu, Psat, Xass0=None):
        """
        dOm_aux(rhoi, temp_aux, mu, Psat, Xass0)

        Method that computes the Thermodynamic Grand potential
        at given density and temperature.

        Parameters
        ----------
        rhoi : array_like
            density vector [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        mu : float
            dimensionless chemical potential at equilibrium [Adim]
        Psat : float
            equilibrium pressure divided by RT [Pa mol / J]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        dom: float
            Thermodynamic Grand potential divided by RT [Pa mol / J]
        Xass : array
            computed fraction of nonbonded sites
        """
        a0ad, Xass = self.a0ad_aux(rhoi, temp_aux, Xass0)
        dom = a0ad - np.sum(np.nan_to_num(rhoi*mu)) + Psat
        return dom, Xass

    def dOm(self, rhoi, T, mu, Psat, Xass0=None):
        """
        dOm(rhoi, T, mu, Psat, Xass0)

        Method that computes the Thermodynamic Grand potential
        at given density and temperature.

        Parameters
        ----------
        rhoi : array_like
            density vector [mol/m^3]
        T : float
            absolute temperature [K]
        mu : float
            dimensionless chemical potential at equilibrium
        Psat : float
            equilibrium pressure divided by RT [Pa mol / J]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        Out: float
            Thermodynamic Grand potential divided by RT [Pa mol / J]
        """
        temp_aux = self.temperature_aux(T)
        dom, Xass = self.dOm_aux(rhoi, temp_aux, mu, Psat, Xass0)
        return dom

    def sgt_adim(self, T):
        '''
        sgt_adim(T)

        Method that evaluates dimensionless factor for temperature, pressure,
        density, tension and distance for interfacial properties computations
        with SGT.

        Parameters
        ----------
        T : float
        absolute temperature [K]

        Returns
        -------
        Tfactor : float
            factor to obtain dimentionless temperature (K -> K)
        Pfactor : float
            factor to obtain dimentionless pressure (Pa -> Pa/RT)
        rofactor : float
            factor to obtain dimentionless density (mol/m3 -> mol/m3)
        tenfactor : float
            factor to obtain surface tension in (mN/m)
        zfactor : float
            factor to obtain the interfacial distance in (Amstrongs)
        '''
        beta = 1 / (kb*T)
        RT = (Na/beta)
        cii0 = np.polyval(self.cii[0], T)  # computing first component cii

        Tfactor = 1.
        Pfactor = 1. / RT
        rofactor = 1.
        tenfactor = np.sqrt(cii0*RT) * 1000  # To give tension in mN/m
        zfactor = 10**-10 * np.sqrt(RT / cii0)
        return Tfactor, Pfactor, rofactor, tenfactor, zfactor

    def beta_sgt(self, beta0, beta1=None, beta2=None, beta3=None):
        r"""
        beta_sgt(beta)

        Method that adds beta correction for cross influence parameters used
        in SGT. The beta correction is computed as follows:

        .. math::
            \beta_{ij} =  \beta_{ij,0} + \beta_{ij,1} \cdot T +  \beta_{ij,2} \cdot T^2 + \frac{\beta_{ij,3}}{T}

        Parameters
        ----------
        beta0 : array_like
            beta0 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [Adim]
        beta1 : array_like, optional
            beta1 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [1/K].
            If None, then a zero matrix is assumed.
        beta2 : array_like, optional
            beta2 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [1/K^2].
            If None, then a zero matrix is assumed.
        beta3 : array_like, optional
            beta3 matrix (Symmetric, Diagonal==0, shape=(nc, nc)) [K]
            If None, then a zero matrix is assumed.

        """
        nc = self.nc

        Beta0 = np.asarray(beta0)
        shape = Beta0.shape
        isSquare = shape == (nc, nc)
        isSymmetric = np.allclose(Beta0, Beta0.T)
        diagZero = np.all(np.diagonal(Beta0) == 0.)
        if isSquare and isSymmetric and diagZero:
            self.beta0 = Beta0
        else:
            raise Exception('beta0 matrix is not square, symmetric or diagonal==0')

        if beta1 is None:
            Beta1 = np.zeros([nc, nc])
            self.beta1 = Beta1
        else:
            Beta1 = np.asarray(beta1)
            shape = Beta1.shape
            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(Beta1, Beta1.T)
            diagZero = np.all(np.diagonal(Beta1) == 0.)
            if isSquare and isSymmetric and diagZero:
                self.beta1 = Beta1
            else:
                raise Exception('beta1 matrix is not square, symmetric or diagonal==0')
        if beta2 is None:
            Beta2 = np.zeros([nc, nc])
            self.beta2 = Beta2
        else:
            Beta2 = np.asarray(beta2)
            shape = Beta2.shape
            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(Beta2, Beta2.T)
            diagZero = np.all(np.diagonal(Beta2) == 0.)
            if isSquare and isSymmetric and diagZero:
                self.beta2 = Beta2
            else:
                raise Exception('beta2 matrix is not square, symmetric or diagonal==0')

        if beta3 is None:
            Beta3 = np.zeros([nc, nc])
            self.beta3 = Beta3
        else:
            Beta3 = np.asarray(beta3)
            shape = Beta3.shape
            isSquare = shape == (nc, nc)
            isSymmetric = np.allclose(Beta3, Beta3.T)
            diagZero = np.all(np.diagonal(Beta3) == 0.)
            if isSquare and isSymmetric and diagZero:
                self.beta3 = Beta3
            else:
                raise Exception('beta3 matrix is not square, symmetric or diagonal==0')

    def set_betaijsgt(self, i, j, beta0, beta1=0., beta2=0., beta3=0.):
        r"""
        set_betaijsgt(i,j, beta0, beta1, beta2, beta3)

        Method that sets the betaij correction for cross influence parameter
        between component i and j.
        The beta correction is computed as follows:

        .. math::
            \beta_{ij} =  \beta_{ij,0} + \beta_{ij,1} \cdot T +  \beta_{ij,2} \cdot T^2 + \frac{\beta_{ij,3}}{T}

        Parameters
        ----------
        i : int
            index of component i.
        j : int
            index of component j.
        beta0 : float
            beta0 value between component i and j [Adim]
        beta1 : float, optional
            beta1 value between component i and j [1/K]. Default to zero.
        beta2 : float, optional
            beta2 value between component i and j [1/K^2]. Default to zero.
        beta3 : float, optional
            beta3 value between component i and j [K]. Default to zero.

        """
        typei = type(i) == int
        typej = type(j) == int

        nc = self.nc
        nc_i = 0 <= i <= (nc - 1)
        nc_j = 0 <= j <= (nc - 1)

        i_j = i != j

        if (not nc_i) or (not nc_j):
            raise Exception('Index i or j bigger than (nc-1)')
        if not i_j:
            raise Exception('Cannot set betaij for i=j')

        if typei and typej and nc_i and nc_j and i_j:
            self.beta0[i, j] = beta0
            self.beta0[j, i] = beta0

            self.beta1[i, j] = beta1
            self.beta1[j, i] = beta1

            self.beta2[i, j] = beta2
            self.beta2[j, i] = beta2

            self.beta3[i, j] = beta3
            self.beta3[j, i] = beta3

    def ci(self, T):
        """
        Method that computes the matrix of cij interaction parameter for SGT at
        given temperature.

        Parameters
        ----------
        T : float
            absolute temperature [K]

        Returns
        -------
        ci : array_like
            influence parameter matrix at given temperature [J m^5 / mol^2]

        """
        n = self.nc
        ci = np.zeros(n)
        for i in range(n):
            ci[i] = np.polyval(self.cii[i], T)
        cij = np.sqrt(np.outer(ci, ci))

        beta = self.beta0 + self.beta1*T + self.beta2*T**2 + self.beta3/T

        cij *= (1 - beta)
        return cij

    def EntropyR(self, x, T, P, state, v0=None, Xass0=None, T_step=0.1):
        """
        EntropyR(x, T, P, state, v0, Xass0, T_step)

        Method that computes the residual entropy of the mixture at given
        temperature and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy

        Returns
        -------
        Sr : float
            residual entropy [J/mol K]

        """

        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho0 = None
        else:
            rho0 = 1./v0

        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)

        v = 1/rho
        rhomolecular = Na * rho
        a, Xass = ares(self, x, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta
        Z = P * v / RT

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = ares(self, x, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = ares(self, x, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = ares(self, x, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = ares(self, x, rhomolecular, temp_aux_2, Xass)

        F = a
        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h

        Sr_TVN = -T*dFdT - F  # residual entropy (TVN) divided by R
        Sr_TPN = Sr_TVN + np.log(Z)  # residual entropy (TPN) divided by R
        Sr_TPN *= R
        return Sr_TPN

    def EnthalpyR(self, x, T, P, state, v0=None, Xass0=None, T_step=0.1):
        """
        EnthalpyR(x, T, P, state, v0, Xass0, T_step)

        Method that computes the residual enthalpy of the mixture at given
        temperature and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy

        Returns
        -------
        Hr : float
            residual enthalpy [J/mol]

        """
        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho0 = None
        else:
            rho0 = 1./v0
        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)

        v = 1/rho
        rhomolecular = Na * rho
        a, Xass = ares(self, x, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta
        Z = P * v / RT

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = ares(self, x, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = ares(self, x, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = ares(self, x, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = ares(self, x, rhomolecular, temp_aux_2, Xass)

        F = a
        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h

        Sr_TVN = -T*dFdT - F  # residual entropy (TVN) divided by R
        Hr_TPN = F + Sr_TVN + Z - 1.  # residual entalphy divided by RT
        Hr_TPN *= RT
        return Hr_TPN

    def CvR(self, x, rho, T, Xass0=None, T_step=0.1):
        """
        CvR(x, rho, T, Xass0, T_step)

        Method that computes the residual isochoric heat capacity of the
        mixture at given density and temperature.

        Parameters
        ----------
        x: array_like
            molar fraction array
        rho : float
            density [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute temperature numerical derivates of Helmholtz
            free energy

        Returns
        -------
        Cv: float
            isochoric heat capacity [J/mol K]
        """
        temp_aux = self.temperature_aux(T)

        rhomolecular = Na * rho

        a, Xass = ares(self, x, rhomolecular, temp_aux, Xass0)

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = ares(self, x, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = ares(self, x, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = ares(self, x, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = ares(self, x, rhomolecular, temp_aux_2, Xass)

        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        d2FdT = (-a_2/12 + 4*a_1/3 - 5*a/2 + 4*a1/3 - a2/12)/h**2

        Cvr_TVN = -T**2*d2FdT - 2*T*dFdT  # residual isochoric heat capacity
        Cvr_TVN *= R
        return Cvr_TVN

    def CpR(self, x, T, P, state, v0=None, Xass0=None, T_step=0.1):
        """
        Cpr(T, P, state, v0, Xass0, T_step)

        Method that computes the residual heat capacity of the mixture at given
        temperature and pressure.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy

        Returns
        -------
        Cp: float
            residual heat capacity [J/mol K]
        """

        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho0 = None
        else:
            rho0 = 1./v0
        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)

        rhomolecular = Na * rho

        d2a, Xass = d2ares_drho(self, x, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = dares_drho(self, x, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = dares_drho(self, x, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = dares_drho(self, x, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = dares_drho(self, x, rhomolecular, temp_aux_2, Xass)

        a = d2a[:2]
        da_drho = a[1] * Na
        d2a_drho = d2a[2] * Na**2

        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        dFdT[1] *= Na

        d2FdT = (-a_2/12 + 4*a_1/3 - 5*a/2 + 4*a1/3 - a2/12) / h**2
        d2FdT[1] *= Na

        dP_dT = RT*(rho**2 * dFdT[1]) + P/T

        dP_drho = 2*rho*da_drho + 2.
        dP_drho += rho**2 * d2a_drho - 1.
        dP_drho *= RT

        dP_dV = -rho**2 * dP_drho
        # residual isochoric heat capacity
        Cvr_TVN = R * (-T**2*d2FdT[0] - 2*T*dFdT[0])
        # residual heat capacity
        Cpr = Cvr_TVN - R - T*dP_dT**2/dP_dV
        return Cpr

    def speed_sound(self, x, T, P, state, v0=None, Xass0=None, T_step=0.1,
                    CvId=3*R/2, CpId=5*R/2):
        """
        speed_sound(x, T, P, state, v0, Xass0, T_step)

        Method that computes the speed of sound of the mixture at given
        temperature and pressure.

        This calculation requires that the molar weight of the fluids has been
        set in the component function.

        By default the ideal gas Cv and Cp are set to 3R/2 and 5R/2, the user
        can supply better values if available.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy
        CvId: float, optional
            Ideal gas isochoric heat capacity, set to 3R/2 by default [J/mol K]
        CpId: float, optional
            Ideal gas heat capacity, set to 3R/2 by default [J/mol K]

        Returns
        -------
        w: float
            speed of sound [m/s]
        """

        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho0 = None
        else:
            rho0 = 1./v0
        rho, Xass = self.density_aux(x, temp_aux, P, state, rho0, Xass0)

        rhomolecular = Na * rho

        d2a, Xass = d2ares_drho(self, x, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = dares_drho(self, x, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = dares_drho(self, x, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = dares_drho(self, x, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = dares_drho(self, x, rhomolecular, temp_aux_2, Xass)

        a = d2a[:2]
        da_drho = a[1] * Na
        d2a_drho = d2a[2] * Na**2

        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        dFdT[1] *= Na

        d2FdT = (-a_2/12 + 4*a_1/3 - 5*a/2 + 4*a1/3 - a2/12) / h**2
        d2FdT[1] *= Na

        dP_dT = RT*(rho**2 * dFdT[1]) + P/T

        dP_drho = 2*rho*da_drho + 2.
        dP_drho += rho**2 * d2a_drho - 1.
        dP_drho *= RT

        dP_dV = -rho**2 * dP_drho
        # residual isochoric heat capacity
        Cvr_TVN = R * (-T**2*d2FdT[0] - 2*T*dFdT[0])
        # residual heat capacity
        Cpr = Cvr_TVN - R - T*dP_dT**2/dP_dV

        # speed of sound calculation
        Cp = CpId + Cpr
        Cv = CvId + Cvr_TVN

        betas = -rho * (Cv/Cp) / dP_dV

        Mwx = np.dot(x, self.Mw)
        w2 = 1000./(rho * betas * Mwx)
        w = np.sqrt(w2)

        return w

    def get_lnphi_pure(self, T, P, state):
        """
        get_lnphi_pure(T, P, state)

        Method that computes the logarithm of the pure component's fugacity
        coefficient at given state, temperature T and pressure P.

        Parameters
        ----------
        T: float
            absolute temperature [K]
        P: float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapor phase

        Returns
        -------
        lnphi_pure: float
            logarithm of pure component's fugacity coefficient
        """

        lnphi_pure = np.zeros(self.nc)
        for i, pure_eos in enumerate(self.pure_eos):
            lnphi_pure[i], _ = pure_eos.logfug(T, P, state)
        return lnphi_pure

    def get_lngamma(self, x, T, P, v0=None, Xass0=None, lnphi_pure=None):
        """
        get_lngamma(x, T, P, v0, Xass0)

        Method that computes the activity coefficient of the mixture at given
        composition x, temperature T and pressure P.

        Parameters
        ----------
        x: array_like
            molar fraction array
        T: float
            absolute temperature [K]
        P: float
            pressure [Pa]
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        lnphi_pure: array, optional
            logarithm of the pure components's fugacity coefficient.
            Computed if not provided.

        Returns
        -------
        lngamma: float
            logarithm of activity coefficient model
        """

        if isinstance(x, (float, int)):
            x = np.array([x, 1.-x])
        elif not isinstance(x, np.ndarray):
            x = np.array(x)

        if self.nc > 2 and x.shape[0] < 2:
            raise ValueError('Please supply the whole molfrac vector for non-binary mixtures')

        lnphi_mix, _ = self.logfugef(x, T, P, 'L', v0=v0, Xass0=Xass0)

        if lnphi_pure is None:
            lnphi_pure = np.zeros_like(x)
            for i, pure_eos in enumerate(self.pure_eos):
                lnphi_pure[i], _ = pure_eos.logfug(T, P, 'L')

        lngamma = lnphi_mix - lnphi_pure
        return lngamma




from __future__ import division, print_function, absolute_import


class TensionResult(dict):

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):

        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in self.items()])
        else:
            return self.__class__.__name__ + "()"



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import minimize
from copy import copy


def tpd(X, state, Z, T, P, model, v0=[None, None]):
    """
    Michelsen's Adimentional tangent plane function

    tpd (x, state, z, T, P, model)


    Parameters
    ----------
    X : array_like
        mole fraction array of trial fase
    state : string
        'L' for liquid phase, 'V' for vapour phase
    Z : array_like
        mole fraction array of overall mixture
    T :  float
        absolute temperature [K]
    P:  float
        absolute pressure [Pa]
    model : object
        created from mixture and saftvrmie function
    v0 : list, optional
        initial volume to solve fugacity, if supplied

    Returns
    -------
    tpd: float
        tpd distance

    """
    v1, v2 = v0
    logfugX, v1 = model.logfugef(X, T, P, state, v1)
    logfugZ, v2 = model.logfugef(Z, T, P, 'L', v2)
    di = np.log(Z) + logfugZ
    tpdi = X*(np.log(X) + logfugX - di)
    return np.sum(np.nan_to_num(tpdi))


def tpd_obj(alpha, temp_aux, P, di, z_notzero, model, state):
    nc = model.nc
    W = np.zeros(nc)
    W[z_notzero] = alpha**2/4.  # change from alpha to mole numbers
    w = W/W.sum()  # change to mole fraction
    global vgw
    out = model.logfugef_aux(w, temp_aux, P, state, vgw)
    logfugW, vgw, Xassgw = out

    dtpd = np.log(W[z_notzero]) + logfugW[z_notzero] - di
    tpdi = np.nan_to_num(W[z_notzero]*(dtpd-1.))
    tpd = 1. + np.sum(tpdi)
    dtpd *= alpha/2
    return tpd, dtpd


def tpd_min(W, Z, T, P, model, stateW, stateZ, vw=None, vz=None, Xassw=None,
            Xassz=None):
    """

    Found a minimiun of Michelsen's Adimentional tangent plane function

    tpd_min (W, Z, T, P, model, stateW, stateZ)

    Parameters
    ----------
    W : array_like
        mole fraction array of trial fase
    Z : array_like
        mole fraction array of overall mixture
    T : float
        absolute temperature [K]
    P:  float
        absolute pressure [Pa]
    model : object create from mixture, eos and mixrule
    stateW : string
        'L' for liquid phase, 'V' for vapor phase
    stateZ : string
        'L' for liquid phase, 'V' for vapor phase
    vw, vz: float, optional
        initial volume value to compute fugacities of phases
    Xassw, Xassz : float, optional
        if supplied used to computed association nonboned sites fraction

    Returns
    -------
    w : array_like
        molar fraction of minimum
    f : float
        minimized tpd distance

    """

    nc = model.nc
    if len(W) != nc or len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')
    # Fugacity of phase Zl

    temp_aux = model.temperature_aux(T)
    Z = Z.copy()

    Z[Z < 1e-8] = 0.0
    z_notzero = np.nonzero(Z)
    logfugZ, vz, Xassz = model.logfugef_aux(Z, temp_aux, P, stateZ, vz)
    di = np.log(Z[z_notzero]) + logfugZ[z_notzero]

    alpha0 = 2*W**0.5
    alpha0[alpha0 < 1e-2] = 1e-2
    alpha0 = alpha0[z_notzero]

    global vgw
    vgw = copy(vw)

    alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                     args=(temp_aux, P, di, z_notzero, model, stateW))
    W = np.zeros(nc)
    W[z_notzero] = alpha.x**2/4.
    w = W/W.sum()
    tpd = alpha.fun

    '''
    Z[Z < 1e-8] = 1e-8
    logfugZ, vz, Xassz = model.logfugef_aux(Z, temp_aux, P, stateZ, vz)
    di = np.log(Z) + logfugZ

    alpha0 = 2*W**0.5
    alpha0[alpha0 < 1e-8] = 1e-8  # To avoid negative compositions
    global vgw  # , Xassgw
    vgw = copy(vw)
    # Xassgw = copy(Xassw)
    alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                     args=(temp_aux, P, di, model, stateW))

    W = alpha.x**2/2
    w = W/W.sum()
    tpd = alpha.fun
    '''
    return w, tpd


def tpd_minimas(nmin, Z, T, P, model, stateW, stateZ, vw=None, vz=None,
                Xassw=None, Xassz=None):
    """

    Found nmin minimuns of Michelsen's Adimentional tangent plane function

    tpd_minimas (nmin, Z, T, P, model, stateW, stateZ)

    Parameters
    ----------
    nmin: int
        number of minimiuns to be founded
    Z : array_like
        mole fraction array of overall mixture
    T : float
        absolute temperature [K]
    P:  float
        absolute pressure [Pa]
    model : object
        create from mixture, eos and mixrule
    stateW : string
        'L' for liquid phase, 'V' for vapour phase
    stateZ : string
        'L' for liquid phase, 'V' for vapour phase
    vw, vz : float, optional
        if supplied volume used as initial value to compute fugacities
    Xassw, Xassz : float, optional
        if supplied used to computed association nonboned sites fraction

    Returns
    -------
    all_minima: tuple
        molar fractions arrays of minimums
    f_minima: tuple
        minimized tpd distance

    """
    nc = model.nc
    if len(Z) != nc:
        raise Exception('Composition vector lenght must be equal to nc')

    temp_aux = model.temperature_aux(T)
    Z = Z.copy()
    Z[Z < 1e-8] = 0.0
    z_notzero = np.nonzero(Z)
    logfugZ, vz, Xassz = model.logfugef_aux(Z, temp_aux, P, stateZ, vz, Xassz)
    di = np.log(Z[z_notzero]) + logfugZ[z_notzero]

    nc = model.nc
    all_minima = []
    f_minima = []

    # search from pures
    Id = np.eye(nc)
    alpha0 = 2*Id[0]**0.5
    alpha0[alpha0 < 1e-1] = 1e-1  # no negative or zero compositions
    alpha0 = alpha0[z_notzero]

    global vgw
    vgw = copy(vw)

    alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                     args=(temp_aux, P, di, z_notzero, model, stateW))

    W = np.zeros(nc)
    W[z_notzero] = alpha.x**2/4.
    w = W/W.sum()
    tpd = alpha.fun

    all_minima.append(w)
    f_minima.append(tpd)
    for i in range(1, nc):
        # if len(f_minima) >= nmin: break 

        alpha0 = 2*Id[i]**0.5
        alpha0[alpha0 < 1e-1] = 1e-1
        alpha0 = alpha0[z_notzero]
        # vw = None
        vgw = copy(vw)
        # Xassgw = copy(Xassw)
        alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                         args=(temp_aux, P, di, z_notzero, model, stateW))
        W = np.zeros(nc)
        W[z_notzero] = alpha.x**2/4.
        w = W/W.sum()
        tpd = alpha.fun
        if alpha.success:
            add = np.any(np.all(np.isclose(all_minima, w, atol=1e-3), axis=1))
            if not add:
                f_minima.append(tpd)
                all_minima.append(w)
                if len(f_minima) == nmin:
                    break
                    # return tuple(all_minima), np.array(f_minima)


    # random seach
    niter = 0
    while len(f_minima) < nmin and niter < (nmin+1):
        niter += 1
        Al = np.random.rand(nc)
        Al = Al/np.sum(Al)
        alpha0 = 2*Al**0.5
        alpha0[alpha0 < 1e-1] = 1e-1
        alpha0 = alpha0[z_notzero]
        # vw = None
        vgw = copy(vw)
        # Xassgw = copy(Xassw)
        alpha = minimize(tpd_obj, alpha0, jac=True, method='BFGS',
                         args=(temp_aux, P, di, z_notzero, model, stateW))
        W = np.zeros(nc)
        W[z_notzero] = alpha.x**2/4.
        w = W/W.sum()
        tpd = alpha.fun
        if alpha.success:
            add = np.any(np.all(np.isclose(all_minima, w, atol=1e-3), axis=1))
            if not add:
                f_minima.append(tpd)
                all_minima.append(w)
                # if len(f_minima) == nmin:
                    # break
                    # return tuple(all_minima), np.array(f_minima)

    f_minima = np.array(f_minima)
    sort = np.argsort(f_minima)
    f_minima = f_minima[sort]

    all_minima = np.array(all_minima)
    all_minima = all_minima[sort]

    f_minima = list(f_minima)
    all_minima = list(all_minima)

    while len(f_minima) < nmin:
        all_minima.append(all_minima[0])
        f_minima.append(f_minima[0])

    if len(f_minima) > nmin:
        f_minima = f_minima[:nmin]
        all_minima = all_minima[:nmin]

    return tuple(all_minima), np.array(f_minima)


def lle_init(Z, T, P, model, vw=None, vz=None):
    """
    Minimize tpd function to initiate ELL at fixed T and P.

    Parameters
    ----------
    z : array_like
        overall molar fraction array
    T : float
        absolute temperature [K]
    P : float
        absolute pressure [Pa]
    model : object
        created from eos and mixture
    vw, vz : float, optional
        if supplied volume used as initial value to compute fugacities

    Returns
    -------
    x0s: tuple
        Contains two mol fractions arrays
    """
    x0s, tpd0 = tpd_minimas(2, Z, T, P, model, 'L', 'L', vw, vz)
    return x0s


def gmix(X, T, P, state, lnphi0, model, v0=None):
    lnphi, v = model.logfugmix(X, T, P, state, v0)
    gmix = lnphi
    gmix -= np.sum(X*lnphi0)
    gmix += np.sum(np.nan_to_num(X*np.log(X)))
    return gmix



from setuptools import setup, Extension, find_packages
import numpy

try:
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
else:
    use_cython = True

ext_modules = []
if use_cython:
    ext_modules = cythonize([
        Extension('sgtpy.coloc_cy', ['sgtpy/src/coloc_cy.pyx'],
                  include_dirs=[numpy.get_include()]),
        Extension('sgtpy.sgt.cijmix_cy', ['sgtpy/src/cijmix_cy.pyx'],
                  include_dirs=[numpy.get_include()])
    ])
else:
    ext_modules = [
        Extension('sgtpy.coloc_cy', ['sgtpy/src/coloc_cy.c'],
                  include_dirs=[numpy.get_include()]),
        Extension('sgtpy.sgt.cijmix_cy', ['sgtpy/src/cijmix_cy.c'],
                  include_dirs=[numpy.get_include()])
    ]

setup(
    packages=find_packages(include=['sgtpy', 'sgtpy.*']),
    ext_modules=ext_modules,
    include_package_data=True,
    zip_safe=False
)



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root, minimize_scalar
from ..constants import kb, Na
from .EquilibriumResult import EquilibriumResult
from warnings import warn

R = Na * kb


def mu_obj(rho, temp_aux, saft):
    rhol, rhov = Na * rho

    global Xassl, Xassv
    dal, Xassl = saft.d2afcn_aux(rhol, temp_aux, Xassl)
    afcnl, dafcnl, d2afcnl = dal
    Pl = rhol**2 * dafcnl / Na
    dPl = (2 * rhol * dafcnl + rhol**2 * d2afcnl)

    mul = afcnl + rhol*dafcnl
    dmul = Na * (rhol*d2afcnl + 2*dafcnl)

    dav, Xassv = saft.d2afcn_aux(rhov, temp_aux)
    afcnv, dafcnv, d2afcnv = dav
    Pv = rhov**2 * dafcnv / Na
    dPv = (2 * rhov * dafcnv + rhov**2 * d2afcnv)

    muv = afcnv + rhov * dafcnv
    dmuv = Na * (rhol*d2afcnv + 2*dafcnv)

    FO = np.array([mul-muv, Pl - Pv])
    dFO = np.array([[dmul, -dmuv],
                   [dPl, - dPv]])
    return FO, dFO


# objective functions for Pmin and Pmax initiation method
def fobj_pmax(rho, temp_aux, self):
    return - self.pressure_aux(rho, temp_aux)[0]


def fobj_pmin(rho, temp_aux, self):
    return self.pressure_aux(rho, temp_aux)[0]


def psat(saft, T, P0=None, v0=[None, None], Xass0=[None, None],
         full_output=True):

    if saft.critical:
        if T >= saft.Tc:
            warn('Temperature is greater than critical temperature, returning critical point')
            if full_output:
                dict = {'T': saft.Tc, 'P': saft.Pc, 'vl': 1./saft.rhoc,
                        'vv': 1./saft.rhoc, 'Xassl': Xass0[0],
                        'Xassv': Xass0[1], 'success': False, 'iterations': 0}
                out = EquilibriumResult(dict)
            else:
                out = saft.Pc, 1./saft.rhoc, 1./saft.rhoc
            return out

    temp_aux = saft.temperature_aux(T)
    beta = temp_aux[0]
    RT = Na/beta

    global Xassl, Xassv
    Xassl, Xassv = Xass0

    vl, vv = v0
    P0input = P0 is None
    vl0input = vl is None
    vv0input = vv is None
    v0input = not vl0input and not vv0input

    init_method = None
    if saft.critical:
        Tr = T/saft.Tc
        if Tr <= 0.8 and P0input:
            if not v0input:
                init_method = 'zero-pressure'

        elif 0.8 < Tr < 1. and P0input:
            if not v0input:
                init_method = 'pmin-pmax'

    if init_method is None:
        if not P0input:
            good_initial = False
            P = P0
        elif v0input:
            good_initial = True
        else:
            raise Exception('You need to provide either initial pressure or both volumes')
    elif init_method == 'zero-pressure':
        rholP0, XassP0 = saft.density_aux(temp_aux, 0., 'L')
        aresP0, XassP0 = saft.ares(rholP0*Na, T, XassP0)
        logfugl0 = aresP0 - 1. + np.log(RT*rholP0)
        fugl0 = np.exp(logfugl0)
        P = fugl0
        vl = 1./rholP0
        Xassl = XassP0
        good_initial = False
    elif init_method == 'pmin-pmax':
        limits_rhov = [saft.rhoc*1e-10, saft.rhoc]
        minpmax = minimize_scalar(fobj_pmax, bounds=limits_rhov,
                                  args=(temp_aux, saft), method='bounded')
        limits_rhol = [saft.rhoc, 5*saft.rhoc]
        minpmin = minimize_scalar(fobj_pmin, bounds=limits_rhol,
                                  args=(temp_aux, saft), method='bounded')
        P = (np.max([0, minpmin.fun]) - minpmax.fun) / 2.
        good_initial = False

    if not good_initial:
        lnphiv, vv, Xassv = saft.logfug_aux(temp_aux, P, 'V', vv, Xassv)
        lnphil, vl, Xassl = saft.logfug_aux(temp_aux, P, 'L', vl, Xassl)
        FO = lnphiv - lnphil
        dFO = (vv - vl)/RT
        dP = FO/dFO
        if dP > P:
            dP /= 2
        P -= dP
        for i in range(15):
            lnphiv, vv, Xassv = saft.logfug_aux(temp_aux, P, 'V', vv, Xassv)
            lnphil, vl, Xassl = saft.logfug_aux(temp_aux, P, 'L', vl, Xassl)
            FO = lnphiv - lnphil
            dFO = (vv - vl)/RT
            P -= FO/dFO
            success = abs(FO) <= 1e-8
            if success:
                break
        if not success:
            rho0 = 1. / np.array([vl, vv])
            sol = root(mu_obj, rho0, args=(temp_aux, saft), jac=True)
            sucess = sol.success
            i += sol.nfev
            rhol, rhov = sol.x
            vl, vv = 1./sol.x
            rhomolecular = rhol * Na
            dal, Xassl = saft.dafcn_aux(rhomolecular, temp_aux, Xassl)
            afcn, dafcn = dal
            P = rhomolecular**2 * dafcn/Na
    else:
        rho0 = 1. / np.asarray([v0])
        sol = root(mu_obj, rho0, args=(temp_aux, saft), jac=True)
        success = sol.success
        i = sol.nfev
        if sol.success:
            rhol, rhov = sol.x
            vl, vv = 1./sol.x
            rhomolecular = rhol * Na
            dal, Xassl = saft.dafcn_aux(rhomolecular, temp_aux, Xassl)
            afcn, dafcn = dal
            P = rhomolecular**2 * dafcn/Na
        else:
            P = None

    if full_output:
        dict = {'T': T, 'P': P, 'vl': vl, 'vv': vv, 'Xassl': Xassl,
                'Xassv': Xassv, 'success': success, 'iterations': i}
        out = EquilibriumResult(dict)
    else:
        out = P, vl, vv
    return out



from __future__ import division, print_function, absolute_import
import numpy as np


# Third pertubation
def a3(xs, xhim, epsij, f4, f5, f6):
    a3ij = np.exp(f5 * xhim + f6 * xhim**2)
    a3ij *= -epsij**3 * f4 * xhim
    a = np.dot(xs, np.dot(a3ij, xs))
    return a


def da3_dxhi00(xs, xhim, dxhim_dxhi00, epsij, f4, f5, f6):

    cte = -epsij**3 * f4 * np.exp(f5 * xhim + f6 * xhim**2)

    a3ij = cte * xhim

    da3ij = cte * (1 + f5 * xhim + 2 * f6 * xhim**2) * dxhim_dxhi00

    a3 = np.array([a3ij, da3ij])
    a = np.matmul(np.matmul(a3, xs), xs)

    return a


def d2a3_dxhi00(xs, xhim, dxhim_dxhi00, epsij, f4, f5, f6):

    cte = -epsij**3 * f4 * np.exp(f5 * xhim + f6 * xhim**2)

    a3ij = cte * xhim

    da3ij = cte * (1 + f5 * xhim + 2 * f6 * xhim**2) * dxhim_dxhi00

    d2a3ij = cte * dxhim_dxhi00**2
    aux3 = 2 * f5 + f5**2 * xhim + 6 * f6 * xhim
    aux3 += 4 * f5 * f6 * xhim**2 + 4 * f6**2*xhim**3
    d2a3ij *= aux3
    a3 = np.array([a3ij, da3ij, d2a3ij])
    a = np.matmul(np.matmul(a3, xs), xs)

    return a


def da3_dx(xs, dxs_dx, xhim, dxhim_dx, epsij, f4, f5, f6):
    aux1 = -epsij**3 * f4 * np.exp(f5 * xhim + f6 * xhim**2)
    aux2 = (1 + f5 * xhim + 2 * f6 * xhim**2)
    aux3 = aux1 * aux2

    a3ij = aux1 * xhim

    da3ijx = np.multiply.outer(dxhim_dx, aux3)

    aux4 = xs * a3ij
    a3m = np.sum(aux4.T*xs)
    suma1 = 2*np.sum(dxs_dx@aux4, axis=1)

    dax = xs@da3ijx@xs + suma1
    return a3m, dax


def da3_dxxhi(xs, dxs_dx, xhim, dxhim_dxhi00, dxhim_dx, epsij, f4, f5, f6):

    aux1 = -epsij**3 * f4 * np.exp(f5 * xhim + f6 * xhim**2)
    aux2 = (1 + f5 * xhim + 2 * f6 * xhim**2)
    aux3 = aux1 * aux2

    a3ij = aux1 * xhim
    da3ij = aux3 * dxhim_dxhi00
    a3 = np.array([a3ij, da3ij])
    a3m = np.matmul(np.matmul(a3, xs), xs)

    da3ijx = np.multiply.outer(dxhim_dx, aux3)

    aux1 = xs * a3ij
    suma1 = 2*np.sum(dxs_dx@aux1, axis=1)

    dax = xs@da3ijx@xs + suma1

    return a3m, dax



from __future__ import division, print_function, absolute_import
import numpy as np


# Equation (53) Paper 2014
def g1sigma(xhi00, xs_m, da1_chain, suma_g1, a1vdw_cteii):
    g1 = 3 * da1_chain - suma_g1/xhi00
    g1 /= xs_m
    g1 /= -a1vdw_cteii
    return g1


def dg1sigma_dxhi00(xhi00, xs_m, d2a1_chain, dsuma_g1, a1vdw_cteii):

    sum1, dsum1 = dsuma_g1
    g1 = 3 * d2a1_chain
    g1[0] -= sum1/xhi00
    g1[1] += (sum1/xhi00 - dsum1)/xhi00
    g1 /= xs_m
    g1 /= -a1vdw_cteii

    return g1


def d2g1sigma_dxhi00(xhi00, xs_m, d3a1_chain, d2suma_g1, a1vdw_cteii):
    sum1, dsum1, d2sum1 = d2suma_g1
    g1 = 3. * d3a1_chain - d2suma_g1/xhi00
    g1[1] += sum1/xhi00**2
    g1[2] += -2*sum1/xhi00**3 + 2*dsum1/xhi00**2
    g1 /= xs_m
    g1 /= -a1vdw_cteii
    return g1



from __future__ import division, print_function, absolute_import
import numpy as np
from .a1sB_monomer import a1sB_eval, da1sB_dxhi00_eval, d2a1sB_dxhi00_eval
from .a1sB_monomer import d3a1sB_dxhi00_eval
from .ahs_monomer import ahs, dahs_dxhi00, d2ahs_dxhi00
from .a3_monomer import a3, da3_dxhi00, d2a3_dxhi00
from .a2_monomer import a2, da2_dxhi00, d2a2_dxhi00
from .a1_monomer import a1, da1_dxhi00, d2a1_dxhi00
from .monomer_aux import kHS, dkHS_dxhi00, d2kHS_dxhi00, d3kHS_dxhi00
from .a2new_chain import da2new_dxhi00, d2a2new_dxhi00, d3a2new_dxhi00
from .gdHS_chain import gdHS, dgdHS_dxhi00, d2gdHS_dxhi00
from .gammac_chain import gammac, dgammac_dxhi00, d2gammac_dxhi00
from .g1sigma_chain import g1sigma, dg1sigma_dxhi00, d2g1sigma_dxhi00
from .g2mca_chain import g2mca, dg2mca_dxhi00, d2g2mca_dxhi00
from .lngmie_chain import lngmie, dlngmie_dxhi00, d2lngmie_dxhi00
from .association_aux import Xass_solver, CIJ_matrix, dXass_drho, d2Xass_drho
from .association_aux import Iab, dIab_drho, d2Iab_drho


# Eq. (14) Paper 2014
def xhi_eval(xhi00, xs_k, xs_m, d_kk03):
    dxhi_dxhi00 = xs_m * np.matmul(xs_k, d_kk03)
    xhi = xhi00 * dxhi_dxhi00
    return xhi, dxhi_dxhi00


# Eq (22) Paper 2014
def xhix_eval(xhi00, xs_k, xs_m, d_kl3):
    sum1 = np.matmul(np.matmul(xs_k, d_kl3), xs_k)
    dxhix_dxhi00 = xs_m * sum1
    xhix = xhi00 * dxhix_dxhi00
    return xhix, dxhix_dxhi00


def ares(self, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3
    xs_k = self.xs_k
    xs_m = self.xs_m
    Ckl = self.Ckl
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    ccteskl = self.ccteskl

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3]])

    # monomer calculation
    a1kl, a2kl = a1sB_eval(xhi00, xhix, xhix_vec[0], xs_m, I_lambdaskl,
                           J_lambdaskl, ccteskl, a1vdwkl, a1vdw_ctekl)

    # zero order pertubation
    aHS = ahs(xhi)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(a1kl * x0_a1, axis=0)
    a1m = a1(xs_k, suma1_monomer)

    # second order pertubation
    khs = kHS(xhix)
    suma2_monomer = Ckl**2 * np.sum(a2kl * x0_a2, axis=0)
    a2m = a2(xs_k, khs, xhixm, suma2_monomer, eps_kl, f1, f2, f3)

    # third order pertubaton
    a3m = a3(xs_k, xhixm, eps_kl, f4, f5, f6)

    am = xs_m * (aHS + beta * a1m + beta2 * a2m + beta3 * a3m)

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    da1ii, da2ii = da1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m,
                                     I_lambdasii, J_lambdasii, cctesii,
                                     a1vdwii, a1vdw_cteii, dxhix_dxhi00)

    # g hard sphere
    gHS = gdHS(x0i_matrix, xhix)

    # gamma_c
    gc = gammac(xhixm, alphaii, tetha)

    # g1sigma
    da1_chain = Cii * np.sum(da1ii[1] * x0_a1ii, axis=0)
    suma_g1 = Cii * np.sum(da1ii[0] * x0_g1ii, axis=0)
    g1s = g1sigma(xhi00, xs_m, da1_chain, suma_g1, a1vdw_cteii)

    # g2sigma
    khs, dkhs = dkHS_dxhi00(xhix, dxhix_dxhi00)
    dsuma2_chain = Cii2 * np.sum(da2ii * x0_a2ii, axis=1)
    da2new = da2new_dxhi00(khs, dkhs, dsuma2_chain, eps_ii)
    suma_g2 = Cii2 * np.sum(da2ii[0] * x0_g2ii, axis=0)
    g2m = g2mca(xhi00, khs, xs_m, da2new, suma_g2, eps_ii, a1vdw_cteii)
    g2s = (1 + gc) * g2m

    lng = lngmie(gHS, g1s, g2s, beps_ii, beps_ii2)
    ach = - (xs_m - 1)*lng

    ares = am + ach

    if self.asso_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        T_ad = temp_aux[29]
        Fklab = temp_aux[30]
        sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
        rho_ad = rho * xs_m * sigma_x3

        Iijklab = Iab(rho_ad, T_ad)

        diagasso = self.diagasso
        vki_asso = self.vki[self.group_asso_index]
        DIJ = self.DIJ
        Dijklab = self.kAB_kl * Fklab * Iijklab

        Xass = Xass_solver(rho, vki_asso, DIJ, Dijklab, diagasso, Xass)
        ares += np.dot(self.S * vki_asso, (np.log(Xass) - Xass/2 + 1/2))
    else:
        Xass = None
    return ares, Xass


def dares_drho(self, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3
    xs_k = self.xs_k
    xs_m = self.xs_m
    Ckl = self.Ckl
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2., 6*xhix, 12*xhix2]])

    # monomer contribution calculation
    da1kl, da2kl = da1sB_dxhi00_eval(xhi00, xhix, xhix_vec[:2], xs_m,
                                     I_lambdaskl, J_lambdaskl, ccteskl,
                                     a1vdwkl, a1vdw_ctekl, dxhix_dxhi00)

    # zero order pertubation
    daHS = dahs_dxhi00(xhi, dxhi_dxhi00)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(da1kl * x0_a1, axis=1)
    da1m = da1_dxhi00(xs_k, suma1_monomer)

    # second order pertubation
    khs, dkhs = dkHS_dxhi00(xhix, dxhix_dxhi00)
    suma2_monomer = Ckl**2 * np.sum(da2kl * x0_a2, axis=1)
    da2m = da2_dxhi00(xs_k, khs, dkhs, xhixm, dxhixm_dxhi00, suma2_monomer,
                      eps_kl, f1, f2, f3)

    # third order pertubaton
    da3m = da3_dxhi00(xs_k, xhixm, dxhixm_dxhi00, eps_kl, f4, f5, f6)

    damono = xs_m * (daHS + beta * da1m + beta2 * da2m + beta3 * da3m)

    # chain contribution calculation
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    d2a1ii, d2a2ii = d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m,
                                        I_lambdasii, J_lambdasii, cctesii,
                                        a1vdwii, a1vdw_cteii, dxhix_dxhi00)

    # g hard sphere
    dgHS = dgdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00)

    # gamma_c
    dgc = dgammac_dxhi00(xhixm, dxhixm_dxhi00, alphaii, tetha)

    # g1sigma
    d2a1_chain = Cii * np.sum(d2a1ii[1:] * x0_a1ii, axis=1)
    dsuma_g1 = Cii * np.sum(d2a1ii[:2] * x0_g1ii, axis=1)
    dg1s = dg1sigma_dxhi00(xhi00, xs_m, d2a1_chain, dsuma_g1, a1vdw_cteii)

    # g2sigma
    khs, dkhs, d2khs = d2kHS_dxhi00(xhix, dxhix_dxhi00)
    d2suma2_chain = Cii2 * np.sum(d2a2ii * x0_a2ii, axis=1)
    d2a2new = d2a2new_dxhi00(khs, dkhs, d2khs, d2suma2_chain, eps_ii)
    dsuma_g2 = Cii2 * np.sum(d2a2ii[:2] * x0_g2ii, axis=1)
    dg2m = dg2mca_dxhi00(xhi00, khs, dkhs, xs_m, d2a2new, dsuma_g2, eps_ii,
                         a1vdw_cteii)
    dg2s = dg2m * (1 + dgc[0])
    dg2s[1] += dg2m[0] * dgc[1]

    dlng = dlngmie_dxhi00(dgHS, dg1s, dg2s, beps_ii, beps_ii2)

    dachain = - (xs_m - 1.)*dlng

    ares = damono + dachain
    ares *= self.dxhi00_1

    if self.asso_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        T_ad = temp_aux[29]
        Fklab = temp_aux[30]
        sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
        drho_ad = xs_m * sigma_x3
        rho_ad = rho * drho_ad

        Iijklab, dIijklab_drho = dIab_drho(rho_ad, T_ad, drho_ad)

        diagasso = self.diagasso
        vki_asso = self.vki[self.group_asso_index]
        DIJ = self.DIJ
        Dijklab = self.kAB_kl * Fklab * Iijklab
        dDijklab_drho = self.kAB_kl * Fklab * dIijklab_drho

        Xass = Xass_solver(rho, vki_asso, DIJ, Dijklab, diagasso, Xass)
        CIJ = CIJ_matrix(rho, vki_asso, Xass, DIJ, Dijklab, diagasso)
        dXass = dXass_drho(rho, vki_asso, Xass, DIJ, Dijklab, dDijklab_drho,
                           CIJ)

        Svki = self.S * vki_asso
        ares[0] += np.dot(Svki, (np.log(Xass) - Xass/2 + 1/2))
        ares[1] += np.dot(Svki, (1/Xass - 1/2) * dXass)
    else:
        Xass = None
    return ares, Xass


def d2ares_drho(self, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3
    xs_k = self.xs_k
    xs_m = self.xs_m
    Ckl = self.Ckl
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3.*xhix2, 4.*xhix3],
                         [0., 2., 6*xhix, 12.*xhix2],
                         [0., 0., 6., 24.*xhix]])

    # monomer contribution calculation
    d2a1kl, d2a2kl = d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec[:3], xs_m,
                                        I_lambdaskl, J_lambdaskl, ccteskl,
                                        a1vdwkl, a1vdw_ctekl, dxhix_dxhi00)

    # zero order pertubation
    d2aHS = d2ahs_dxhi00(xhi, dxhi_dxhi00)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(d2a1kl * x0_a1, axis=1)
    d2a1m = d2a1_dxhi00(xs_k, suma1_monomer)

    # second order pertubation
    khs, dkhs, d2khs = d2kHS_dxhi00(xhix, dxhix_dxhi00)
    suma2_monomer = Ckl**2 * np.sum(d2a2kl * x0_a2, axis=1)
    d2a2m = d2a2_dxhi00(xs_k, khs, dkhs, d2khs, xhixm, dxhixm_dxhi00,
                        suma2_monomer, eps_kl, f1, f2, f3)

    # third order pertubaton
    d2a3m = d2a3_dxhi00(xs_k, xhixm, dxhixm_dxhi00, eps_kl, f4, f5, f6)

    d2amono = xs_m * (d2aHS + beta * d2a1m + beta2 * d2a2m + beta3 * d2a3m)

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    d3a1ii, d3a2ii = d3a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m,
                                        I_lambdasii, J_lambdasii, cctesii,
                                        a1vdwii, a1vdw_cteii, dxhix_dxhi00)

    # g hard sphere
    d2gHS = d2gdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00)

    # gamma_c
    d2gc = d2gammac_dxhi00(xhixm, dxhixm_dxhi00, alphaii, tetha)

    # g1sigma
    d3a1_chain = Cii * np.sum(d3a1ii[1:] * x0_a1ii, axis=1)
    d2suma_g1 = Cii * np.sum(d3a1ii[:3] * x0_g1ii, axis=1)
    d2g1s = d2g1sigma_dxhi00(xhi00, xs_m, d3a1_chain, d2suma_g1, a1vdw_cteii)

    # g2sigma
    khs, dkhs, d2khs, d3khs = d3kHS_dxhi00(xhix, dxhix_dxhi00)
    d3suma2_chain = Cii2 * np.sum(d3a2ii * x0_a2ii, axis=1)
    d3a2new = d3a2new_dxhi00(khs, dkhs, d2khs, d3khs, d3suma2_chain, eps_ii)
    d2suma_g2 = Cii2 * np.sum(d3a2ii[:3] * x0_g2ii, axis=1)
    d2g2m = d2g2mca_dxhi00(xhi00, khs, dkhs, d2khs, xs_m, d3a2new, d2suma_g2,
                           eps_ii, a1vdw_cteii)
    d2g2s = d2g2m * (1. + d2gc[0])
    d2g2s[1] += d2g2m[0] * d2gc[1]
    d2g2s[2] += 2. * d2g2m[1] * d2gc[1] + d2g2m[0] * d2gc[2]

    d2lng = d2lngmie_dxhi00(d2gHS, d2g1s, d2g2s, beps_ii, beps_ii2)

    d2achain = - (xs_m - 1.)*d2lng

    ares = d2amono + d2achain
    ares *= self.dxhi00_2

    if self.asso_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        T_ad = temp_aux[29]
        Fklab = temp_aux[30]
        sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
        drho_ad = xs_m * sigma_x3
        rho_ad = rho * drho_ad

        Iijklab, dIijklab_drho, d2Iijklab_drho = d2Iab_drho(rho_ad, T_ad,
                                                            drho_ad)

        diagasso = self.diagasso
        vki_asso = self.vki[self.group_asso_index]
        DIJ = self.DIJ
        Dijklab = self.kAB_kl * Fklab * Iijklab
        dDijklab_drho = self.kAB_kl * Fklab * dIijklab_drho
        d2Dijklab_drho = self.kAB_kl * Fklab * d2Iijklab_drho

        Xass = Xass_solver(rho, vki_asso, DIJ, Dijklab, diagasso, Xass)
        CIJ = CIJ_matrix(rho, vki_asso, Xass, DIJ, Dijklab, diagasso)
        dXass = dXass_drho(rho, vki_asso, Xass, DIJ, Dijklab, dDijklab_drho,
                           CIJ)
        d2Xass = d2Xass_drho(rho, vki_asso, Xass, dXass, DIJ, Dijklab,
                             dDijklab_drho, d2Dijklab_drho, CIJ)

        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2
        Svki = self.S * vki_asso
        ares[0] += np.dot(Svki, aux1)
        ares[1] += np.dot(Svki, aux2 * dXass)
        ares[2] += np.dot(Svki, -(dXass/Xass)**2+d2Xass*aux2)
    else:
        Xass = None
    return ares, Xass



from __future__ import division, print_function, absolute_import
import numpy as np
from numba import jit
from scipy.optimize import fsolve


cpq = np.zeros([11, 11])
cpq[0, 0] = 7.56425183020431E-02
cpq[0, 1] = -1.28667137050961E-01
cpq[0, 2] = 1.28350632316055E-01
cpq[0, 3] = -7.25321780970292E-02
cpq[0, 4] = 2.57782547511452E-02
cpq[0, 5] = -6.01170055221687E-03
cpq[0, 6] = 9.33363147191978E-04
cpq[0, 7] = -9.55607377143667E-05
cpq[0, 8] = 6.19576039900837E-06
cpq[0, 9] = -2.30466608213628E-07
cpq[0, 10] = 3.74605718435540E-09
cpq[1, 0] = 1.34228218276565E-01
cpq[1, 1] = -1.82682168504886E-01
cpq[1, 2] = 7.71662412959262E-02
cpq[1, 3] = -7.17458641164565E-04
cpq[1, 4] = -8.72427344283170E-03
cpq[1, 5] = 2.97971836051287E-03
cpq[1, 6] = -4.84863997651451E-04
cpq[1, 7] = 4.35262491516424E-05
cpq[1, 8] = -2.07789181640066E-06
cpq[1, 9] = 4.13749349344802E-08
cpq[2, 0] = -5.65116428942893E-01
cpq[2, 1] = 1.00930692226792E+00
cpq[2, 2] = -6.60166945915607E-01
cpq[2, 3] = 2.14492212294301E-01
cpq[2, 4] = -3.88462990166792E-02
cpq[2, 5] = 4.06016982985030E-03
cpq[2, 6] = -2.39515566373142E-04
cpq[2, 7] = 7.25488368831468E-06
cpq[2, 8] = -8.58904640281928E-08
cpq[3, 0] = -3.87336382687019E-01
cpq[3, 1] = -2.11614570109503E-01
cpq[3, 2] = 4.50442894490509E-01
cpq[3, 3] = -1.76931752538907E-01
cpq[3, 4] = 3.17171522104923E-02
cpq[3, 5] = -2.91368915845693E-03
cpq[3, 6] = 1.30193710011706E-04
cpq[3, 7] = -2.14505500786531E-06
cpq[4, 0] = 2.13713180911797E+00
cpq[4, 1] = -2.02798460133021E+00
cpq[4, 2] = 3.36709255682693E-01
cpq[4, 3] = 1.18106507393722E-03
cpq[4, 4] = -6.00058423301506E-03
cpq[4, 5] = 6.26343952584415E-04
cpq[4, 6] = -2.03636395699819E-05
cpq[5, 0] = -3.00527494795524E-01
cpq[5, 1] = 2.89920714512243E+00
cpq[5, 2] = -5.67134839686498E-01
cpq[5, 3] = 5.18085125423494E-02
cpq[5, 4] = -2.39326776760414E-03
cpq[5, 5] = 4.15107362643844E-05
cpq[6, 0] = -6.21028065719194E+00
cpq[6, 1] = -1.92883360342573E+00
cpq[6, 2] = 2.84109761066570E-01
cpq[6, 3] = -1.57606767372364E-02
cpq[6, 4] = 3.68599073256615E-04
cpq[7, 0] = 1.16083532818029E+01
cpq[7, 1] = 7.42215544511197E-01
cpq[7, 2] = -8.23976531246117E-02
cpq[7, 3] = 1.86167650098254E-03
cpq[8, 0] = -1.02632535542427E+01
cpq[8, 1] = -1.25035689035085E-01
cpq[8, 2] = 1.14299144831867E-02
cpq[9, 0] = 4.65297446837297E+00
cpq[9, 1] = -1.92518067137033E-03
cpq[10, 0] = -8.67296219639940E-01


@jit(cache=True, nopython=True)
def Iab(rho_ad, T_ad):
    Iijklab = 0.
    for p in range(11):
        for q in range(0, 11-p):
            Iijklab += cpq[p, q] * rho_ad**p * T_ad**q
            pass
    return Iijklab


@jit(cache=True, nopython=True)
def dIab_drho(rho_ad, T_ad, drho_ad):
    Iijklab = 0.
    dIijklab_drho = 0.
    for p in range(11):
        for q in range(0, 11-p):
            Iijklab += cpq[p, q] * rho_ad**p * T_ad**q
            dIijklab_drho += cpq[p, q] * p * rho_ad**(p-1) * T_ad**q
    dIijklab_drho *= drho_ad
    return Iijklab, dIijklab_drho


@jit(cache=True, nopython=True)
def d2Iab_drho(rho_ad, T_ad, drho_ad):
    Iijklab = 0.
    dIijklab_drho = 0.
    d2Iijklab_drho = 0.
    for p in range(11):
        for q in range(0, 11-p):
            Iijklab += cpq[p, q] * rho_ad**p * T_ad**q
            dIijklab_drho += cpq[p, q] * p * rho_ad**(p-1) * T_ad**q
            d2Iijklab_drho += cpq[p, q] * p * (p-1) * rho_ad**(p-2) * T_ad**q
    dIijklab_drho *= drho_ad
    d2Iijklab_drho *= drho_ad**2
    return Iijklab, dIijklab_drho, d2Iijklab_drho


def fobj_xass(Xass, vki_asso, aux_asso, diagasso):
    fo = Xass - 1 / (1 + aux_asso@(vki_asso*Xass))
    return fo


def fobj_xass_jac(Xass, vki_asso, aux_asso, diagasso):
    den = 1 + aux_asso@(vki_asso*Xass)
    dfo = ((aux_asso*vki_asso).T/den**2).T
    dfo[diagasso] += 1.
    return dfo


def Xass_solver(rho, vki_asso, DIJ, Dabij, diagasso, Xass0):

    aux_asso = rho * DIJ * Dabij

    omega = 0.2
    Xass = 1.*Xass0
    for i in range(5):
        den = 1. + aux_asso@(vki_asso*Xass)
        fo = 1. / den
        dXass = (1 - omega) * (fo - Xass)
        Xass += dXass

    KIJ = np.outer(vki_asso, vki_asso) * aux_asso

    KIJXass = KIJ@Xass
    dQ = vki_asso * (1/Xass - 1) - KIJXass
    HIJ = -1 * KIJ
    HIJ[diagasso] -= (vki_asso + KIJXass)/Xass
    for i in range(15):
        dXass = np.linalg.solve(HIJ, -dQ)
        Xnew = Xass + dXass

        is_nan = np.isnan(Xnew)
        Xnew[is_nan] = 0.2

        Xnew_neg = Xnew < 0.
        Xnew[Xnew_neg] = 0.2*Xass[Xnew_neg]

        Xnew_big = Xnew > 1
        Xnew[Xnew_big] = 1.

        Xass = Xnew
        KIJXass = KIJ@Xass
        dQ = vki_asso * (1/Xass - 1) - KIJXass
        sucess = np.linalg.norm(dQ) < 1e-9
        if sucess:
            break
        HIJ = -1 * KIJ
        HIJ[diagasso] -= (vki_asso + KIJXass)/Xass

    if not sucess:
        Xass = fsolve(fobj_xass, x0=Xass, args=(vki_asso, aux_asso, diagasso),
                      fprime=fobj_xass_jac)
    return Xass


def CIJ_matrix(rhom, vki_asso, Xass, DIJ, Dabij, diagasso):
    CIJ = rhom * np.outer(Xass**2, vki_asso) * Dabij * DIJ
    CIJ[diagasso] += 1.
    return CIJ


def dXass_drho(rhom, vki_asso, Xass, DIJ, Dabij, dDabij_drho, CIJ):
    brho = -(DIJ*(Dabij + rhom * dDabij_drho))@(vki_asso * Xass)
    brho *= Xass**2
    dXass = np.linalg.solve(CIJ, brho)
    return dXass


def d2Xass_drho(rhom, vki_asso, Xass, dXass_drho, DIJ, Dabij, dDabij_drho,
                d2Dabij_drho, CIJ):

    b2rho = -rhom * (DIJ * (Xass * d2Dabij_drho+2*dXass_drho*dDabij_drho))@vki_asso
    b2rho += 2 * (1/Xass - 1) / (rhom**2)
    b2rho *= Xass**2
    b2rho += 2 * dXass_drho / (rhom)
    b2rho += 2 * dXass_drho**2 / (Xass)

    d2Xass = np.linalg.solve(CIJ, b2rho)
    return d2Xass


def association_solver(self, rhom, temp_aux, Xass0=None):
    if Xass0 is None:
        Xass = 0.2 * np.ones(self.nsites)
    else:
        Xass = 1. * Xass0

    sigma_kl3 = self.sigma_kl3
    xs_k = self.xs_k
    xs_m = self.xs_m

    T_ad = temp_aux[29]
    Fklab = temp_aux[30]
    sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
    rho_ad = rhom * xs_m * sigma_x3

    Iijklab = Iab(rho_ad, T_ad)

    diagasso = self.diagasso
    vki_asso = self.vki[self.group_asso_index]
    DIJ = self.DIJ
    Dijklab = self.kAB_kl * Fklab * Iijklab

    Xass = Xass_solver(rhom, vki_asso, DIJ, Dijklab, diagasso, Xass)
    return Xass


def association_check(self, rhom, temp_aux, Xass):
    sigma_kl3 = self.sigma_kl3
    xs_k = self.xs_k
    xs_m = self.xs_m

    T_ad = temp_aux[29]
    Fklab = temp_aux[30]
    sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
    rho_ad = rhom * xs_m * sigma_x3

    Iijklab = Iab(rho_ad, T_ad)

    vki_asso = self.vki[self.group_asso_index]
    DIJ = self.DIJ
    Dabij = self.kAB_kl * Fklab * Iijklab

    aux_asso = rhom * DIJ * Dabij
    fo = Xass - 1 / (1 + aux_asso@(vki_asso*Xass))

    return fo



from __future__ import division, print_function, absolute_import
from .vrmie_mixtures import *
from .vrmie_pure import *

from .gammamie_mixtures import *
from .gammamie_pure import *

from .saft import saftvrmie, saftgammamie
from .mixture import component, mixture

from . import equilibrium
#from . import fit
from . import sgt
from .math import *

from .database import df_groups, df_mie_kl, df_asso_kl
from .database import df_secondorder, df_secondasso
from .database import database
